import logging
from pyensembl import genome_for_reference_name
import pandas as pd
from pyensembl import EnsemblRelease
from pyensembl.species import human
from pyensembl.species import Species
from anndata import AnnData

LOGGER = logging.getLogger(__name__)


def map_gene_symbols_to_ensembl_ids(
    adata: AnnData, gene_names: str, species: Species = human
) -> AnnData:
    """
    Map gene names to Ensembl IDs using the pyensembl library.
    Due to copy events, there might be multiple genes per name. We always take the fist one.

    Parameters
    ----------
    adata : anndata.AnnData
        Anndata object containing the gene expression data.
    gene_names : str
        Column name in adata.var containing the gene names.
    species : pyensembl.species.Species, optional, default = human
        Species for which the gene names should be mapped.
        For the provided species, we take the first 'reference_assembly' as default to do the mapping.
        For humans, this is the GRCh38 genome for example.

    Returns
    -------
    anndata.AnnData
        Anndata object with the gene names mapped to Ensembl IDs in adata.var["ensembl_id"]
    """
    # this is one time only
    ensembl_release = EnsemblRelease(species=species)
    ensembl_release.download(overwrite=False)
    ensembl_release.index(overwrite=False)

    adata.var["ensembl_id"] = pd.Series([None] * len(adata.var), index=adata.var.index)
    # we take the first reference assembly from the provided dictionary as default
    genome_reference = genome_for_reference_name(
        next(iter(species.reference_assemblies))
    )
    for index, name in adata.var[gene_names].items():
        try:
            adata.var.at[index, "ensembl_id"] = genome_reference.gene_ids_of_gene_name(
                name
            )[0]
        except:
            continue
    non_none_mappings = adata.var["ensembl_id"].notnull().sum()
    LOGGER.info(
        f"Mapped {non_none_mappings} genes to Ensembl IDs from a total of {adata.var.shape[0]} genes."
    )
    return adata


def map_ensembl_ids_to_gene_symbols(
    adata: AnnData, ensembl_id_key: str = "ensembl_id", species: Species = human
) -> AnnData:
    """
    Map Ensembl IDs to gene names using the pyensembl library.
    We use the GRCh38 genome for mapping.

    Parameters
    ----------
    adata : anndata.AnnData
        Anndata object containing the gene expression data.
    ensembl_id_key : str, optional, default = "ensembl_id"
        Column name in adata.var containing the ensemble ids.
    species : pyensembl.species.Species, optional, default = human
        Species for which the Ensembl IDs should be mapped.
        For the provided species, we take the first reference genome as default to do the mapping.
        For humans, this is the GRCh38 genome for example.

    Returns
    -------
    anndata.AnnData
        Anndata object with the Ensembl IDs mapped to gene names in adata.var["gene_names"]
    """
    # this is one time only
    ensembl_release = EnsemblRelease(species=species)
    ensembl_release.download(overwrite=False)
    ensembl_release.index(overwrite=False)

    adata.var["gene_names"] = pd.Series([None] * len(adata.var), index=adata.var.index)
    # we take the first reference assembly from the provided dictionary as default
    genome_reference = genome_for_reference_name(
        next(iter(species.reference_assemblies))
    )
    for index, ensembl_id in adata.var[ensembl_id_key].items():
        try:
            adata.var.at[index, "gene_names"] = genome_reference.gene_name_of_gene_id(
                ensembl_id
            )
        except:
            continue
    non_none_mappings = adata.var["gene_names"].notnull().sum()
    LOGGER.info(
        f"Mapped {non_none_mappings} genes to Gene names from a total of {adata.var.shape[0]} Ensembl IDs."
    )
    return adata



from typing import Literal, Optional
from helical.models.base_models import (
    HelicalBaseFineTuningHead,
    HelicalBaseFineTuningModel,
)
from helical.models.caduceus import Caduceus, CaduceusConfig
from datasets import Dataset
from transformers import get_scheduler
import torch
from torch import optim
from torch.nn.modules import loss
from tqdm import tqdm
from torch.utils.data import DataLoader
import numpy as np

import logging

LOGGER = logging.getLogger(__name__)


class CaduceusFineTuningModel(HelicalBaseFineTuningModel, Caduceus):
    """CaduceusFineTuningModel
    This model can be used to fine-tune the Caduceus model on a downstream task.

    Example
    ----------
    ```python
    from helical.models.caduceus import CaduceusFineTuningModel, CaduceusConfig

    input_sequences = ["ACT"*20, "ATG"*20, "ATG"*20, "CTG"*20, "TTG"*20]
    labels = [0, 2, 2, 0, 1]

    caduceus_config = CaduceusConfig(model_name="caduceus-ph-4L-seqlen-1k-d118", batch_size=5)
    caduceus_fine_tune = CaduceusFineTuningModel(caduceus_config=caduceus_config, fine_tuning_head="classification", output_size=3)

    train_dataset = caduceus_fine_tune.process_data(input_sequences)

    caduceus_fine_tune.train(train_dataset=train_dataset, train_labels=labels)

    outputs = caduceus_fine_tune.get_outputs(train_dataset)
    print(outputs.shape)
    ```

    Parameters
    ----------
    caduceus_config : CaduceusConfig
        The configuration object for the Caduceus model. The same config object can be used for both the Caduceus and CaduceusFineTuningModel.
    fine_tuning_head : Literal["classification", "regression"] | HelicalBaseFineTuningHead
        The type of fine-tuning head to use for the model. This can be either a classification or regression head, or a custom fine-tuning head.
    output_size : Optional[int]
        The output size of the fine-tuning model. This is required if the fine_tuning_head is a string specified task. For a classification task this is number of unique classes.

    Methods
    ----------
    train(train_dataset, train_labels, optimizer, optimizer_params, loss_function, epochs, freeze_layers, validation_dataset, validation_labels, lr_scheduler_params)
        Fine-tunes the Caduceus model on the given dataset.
    get_outputs(dataset)
        Returns the outputs of the model for the given dataset.
    """

    def __init__(
        self,
        caduceus_config: CaduceusConfig,
        fine_tuning_head: (
            Literal["classification", "regression"] | HelicalBaseFineTuningHead
        ),
        output_size: Optional[int] = None,
    ):
        HelicalBaseFineTuningModel.__init__(self, fine_tuning_head, output_size)
        Caduceus.__init__(self, caduceus_config)

        self.fine_tuning_head.set_dim_size(self.config["embedding_size"])

    def _forward(
        self,
        input_ids: torch.LongTensor = None,
        output_hidden_states: Optional[bool] = None,
        conjoin_train: bool = False,
        conjoin_eval: bool = False,
        training: bool = False,
    ):
        # Get hidden representations from the backbone
        if self.model.config.rcps:  # Hidden states have 2 * d_model channels for RCPS
            transformer_outputs = self.model(
                input_ids=input_ids, output_hidden_states=output_hidden_states
            )
            hidden_states = torch.stack(
                [
                    transformer_outputs[0][..., : self.model.config.d_model],
                    torch.flip(
                        transformer_outputs[0][..., self.model.config.d_model :],
                        dims=[1, 2],
                    ),
                ],
                dim=-1,
            )
        elif conjoin_train or (
            conjoin_eval and not training
        ):  # For conjoining / post-hoc conjoining
            assert input_ids is not None, "`input_ids` must be provided for conjoining."
            assert (
                input_ids.ndim == 3
            ), "`input_ids` must be 3D tensor: channels corresponds to forward and rc strands."
            transformer_outputs = self.model(
                input_ids[..., 0], output_hidden_states=output_hidden_states
            )
            transformer_outputs_rc = self.model(
                input_ids[..., 1], output_hidden_states=output_hidden_states
            )
            # Stack along channel dimension (dim=-1)
            hidden_states = torch.stack(
                [transformer_outputs[0], transformer_outputs_rc[0]], dim=-1
            )
        else:
            transformer_outputs = self.model(
                input_ids, output_hidden_states=output_hidden_states
            )
            hidden_states = transformer_outputs[0]

        # Pool and get logits
        pooled_hidden_states = self._pool_hidden_states(
            hidden_states=hidden_states, sequence_length_dim=1
        )
        # Potentially run `fine_tuning_head` twice (with parameters shared) for conjoining
        if (
            hidden_states.ndim == 4
        ):  # bsz, seq_len, hidden_dim, 2 where last channel has the stacked fwd and rc reps
            logits_fwd = self.fine_tuning_head(pooled_hidden_states[..., 0])
            logits_rc = self.fine_tuning_head(pooled_hidden_states[..., 1])
            logits = (logits_fwd + logits_rc) / 2
        else:
            logits = self.fine_tuning_head(pooled_hidden_states)

        return logits

    def train(
        self,
        train_dataset: Dataset,
        train_labels: np.ndarray,
        optimizer: optim = optim.AdamW,
        optimizer_params: dict = {"lr": 0.0001},
        loss_function: loss = loss.CrossEntropyLoss(),
        epochs: int = 1,
        trainable_layers: int = 2,
        validation_dataset: Optional[Dataset] = None,
        validation_labels: Optional[np.ndarray] = None,
        lr_scheduler_params: Optional[dict] = None,
        conjoin_train: bool = False,
        conjoin_eval: bool = False,
    ):
        """Fine-tunes the Caduceus model on the given dataset.

        Parameters
        ----------
        train_dataset : Dataset
            A helical processed dataset for fine-tuning
        optimizer : torch.optim, default=torch.optim.AdamW
            The optimizer to be used for training.
        optimizer_params : dict, optional, default={'lr': 0.0001}
            The optimizer parameters to be used for the optimizer specified. This list should NOT include model parameters.
            e.g. optimizer_params = {'lr': 0.0001}
        loss_function : torch.nn.modules.loss, default=torch.nn.modules.loss.CrossEntropyLoss()
            The loss function to be used.
        train_labels : np.ndarray
            training labels for the dataset.
        epochs : int, optional, default=10
            The number of epochs to train the model
        trainable_layers : int, optional, default=2
            The number of layers to train in the model. The last n layers will be trained and the rest will be frozen.
        validation_dataset : Dataset, default=None
            A helical processed dataset for per epoch validation. If this is not specified, no validation will be performed.
        lr_scheduler_params : dict, default=None
            The learning rate scheduler parameters for the transformers get_scheduler method. The optimizer will be taken from the optimizer input and should not be included in the learning scheduler parameters. If not specified, no scheduler will be used.
            e.g. lr_scheduler_params = { 'name': 'linear', 'num_warmup_steps': 0 }. num_steps will be calculated based on the number of epochs and the length of the training dataset.
        conjoin_train : bool, default=False
            Whether to conjoin the forward and reverse complement sequences during training.
        conjoin_eval : bool, default=False
            Whether to conjoin the forward and reverse complement sequences during evaluation.
        """
        # initialise optimizer
        optimizer = optimizer(self.parameters(), **optimizer_params)

        # set labels for the dataset
        train_dataset = self._add_data_column(
            train_dataset, "labels", np.array(train_labels)
        )
        if validation_labels is not None and validation_dataset is not None:
            validation_dataset = self._add_data_column(
                validation_dataset, "labels", np.array(validation_labels)
            )

        if trainable_layers > 0:
            LOGGER.info(
                f"Unfreezing the last {trainable_layers} layers of the Caduceus model."
            )

            for param in self.model.backbone.parameters():
                param.requires_grad = False
            for param in self.model.backbone.layers[-trainable_layers:].parameters():
                param.requires_grad = True

        self.to(self.config["device"])

        train_dataloader = DataLoader(
            train_dataset,
            collate_fn=self._collate_fn,
            batch_size=self.config["batch_size"],
            num_workers=self.config["nproc"],
        )

        lr_scheduler = None
        if lr_scheduler_params is not None:
            lr_scheduler = get_scheduler(
                optimizer=optimizer,
                num_training_steps=epochs * len(train_dataloader),
                **lr_scheduler_params,
            )

        if validation_dataset is not None:
            validation_dataloader = DataLoader(
                validation_dataset,
                collate_fn=self._collate_fn,
                batch_size=self.config["batch_size"],
                num_workers=self.config["nproc"],
            )

        LOGGER.info("Starting Fine-Tuning")
        for j in range(epochs):
            training_loop = tqdm(train_dataloader, desc="Fine-Tuning")
            self.model.train()
            self.fine_tuning_head.train()
            batch_loss = 0.0
            batches_processed = 0
            for batch in training_loop:
                input_ids = batch["input_ids"].to(self.config["device"])
                labels = batch["labels"].to(self.config["device"])

                outputs = self._forward(
                    input_ids=input_ids,
                    conjoin_train=conjoin_train,
                    conjoin_eval=conjoin_eval,
                    training=True,
                )

                loss = loss_function(outputs, labels)
                loss.backward()
                optimizer.step()
                optimizer.zero_grad()
                batch_loss += loss.item()
                batches_processed += 1
                training_loop.set_postfix({"loss": batch_loss / batches_processed})
                training_loop.set_description(f"Fine-Tuning: epoch {j+1}/{epochs}")

                del batch
                del outputs

                if lr_scheduler is not None:
                    lr_scheduler.step()

            del training_loop

            if validation_dataset is not None:
                testing_loop = tqdm(
                    validation_dataloader, desc="Fine-Tuning Validation"
                )
                self.model.eval()
                self.fine_tuning_head.eval()
                val_loss = 0.0
                count = 0.0
                for test_batch in testing_loop:
                    input_ids = test_batch["input_ids"].to(self.config["device"])
                    labels = test_batch["labels"].to(self.config["device"])

                    with torch.no_grad():
                        outputs = self._forward(
                            input_ids=input_ids,
                            conjoin_train=conjoin_train,
                            conjoin_eval=conjoin_eval,
                            training=False,
                        )

                    val_loss += loss_function(outputs, labels).item()
                    count += 1.0
                    testing_loop.set_postfix({"val_loss": val_loss / count})

                    del test_batch
                    del outputs

                del testing_loop
                self.model.train()
                self.fine_tuning_head.train()

        LOGGER.info(f"Fine-Tuning Complete. Epochs: {epochs}")

    def get_outputs(self, dataset: Dataset, conjoin: bool = False) -> np.ndarray:
        """Get the embeddings for the tokenized sequence.

        Parameters
        ----------
        dataset : Dataset
            The output dataset from `process_data`.
        conjoin : bool, default=False
            Whether to conjoin the forward and reverse complement sequences.
        """
        dataloader = DataLoader(
            dataset,
            collate_fn=self._collate_fn,
            batch_size=self.config["batch_size"],
            shuffle=False,
            num_workers=self.config["nproc"],
        )
        outputs = []

        self.model.to(self.config["device"])
        self.model.eval()
        self.fine_tuning_head.eval()

        progress_bar = tqdm(dataloader, desc="Generating outputs")
        for batch in progress_bar:
            input_ids = batch["input_ids"].to(self.config["device"])

            with torch.no_grad():
                output = self._forward(
                    input_ids=input_ids, conjoin_eval=conjoin, training=False
                )

            outputs.append(output.cpu().numpy())

            del batch
            del output

        return np.concatenate(outputs)

    def _add_data_column(self, dataset, column_name, data):
        if len(data.shape) > 1:
            for i in range(len(data[0])):  # Assume all inner lists are the same length
                dataset = dataset.add_column(f"{column_name}", [row[i] for row in data])
        else:  # If 1D
            dataset = dataset.add_column(column_name, data)
        return dataset



from typing import Literal, Optional
from helical.models.base_models import (
    HelicalBaseFineTuningHead,
    HelicalBaseFineTuningModel,
)
from helical.models.geneformer import Geneformer, GeneformerConfig
import torch
from torch import optim
from torch.nn.modules import loss
from helical.models.geneformer.geneformer_utils import (
    gen_attention_mask,
    get_model_input_size,
    pad_tensor_list,
    _check_for_expected_special_tokens,
    mean_nonpadding_embs,
)
from datasets import Dataset
import logging
from tqdm import trange
from transformers import get_scheduler


logger = logging.getLogger(__name__)


class GeneformerFineTuningModel(HelicalBaseFineTuningModel, Geneformer):
    """GeneformerFineTuningModel.

    Fine-tuning model for the Geneformer model.

    Example
    ----------
    ```python
    from helical.models.geneformer import GeneformerConfig, GeneformerFineTuningModel
    import anndata as ad

    # Load the data
    ann_data = ad.read_h5ad("/home/matthew/helical-dev/helical/yolksac_human.h5ad")

    # Get the column for fine-tuning
    cell_types = list(ann_data.obs["cell_types"])
    label_set = set(cell_types)

    # Create a GeneformerConfig object
    geneformer_config = GeneformerConfig(model_name="gf-12L-38M-i4096", batch_size=10)

    # Create a GeneformerFineTuningModel object
    geneformer_fine_tune = GeneformerFineTuningModel(geneformer_config=geneformer_config, fine_tuning_head="classification", output_size=len(label_set))

    # Process the data
    dataset = geneformer_fine_tune.process_data(ann_data[:10])

    # Add column to the dataset
    dataset = dataset.add_column('cell_types', cell_types)

    # Create a dictionary to map cell types to ids
    class_id_dict = dict(zip(label_set, [i for i in range(len(label_set))]))

    def classes_to_ids(example):
        example["cell_types"] = class_id_dict[example["cell_types"]]
        return example

    # Convert cell types to ids
    dataset = dataset.map(classes_to_ids, num_proc=1)

    # Fine-tune the model
    geneformer_fine_tune.train(train_dataset=dataset, label="cell_types")

    # Get logits from the fine-tuned model
    outputs = geneformer_fine_tune.get_outputs(dataset)
    print(outputs[:10])

    # Get embeddings from the fine-tuned model
    embeddings = geneformer_fine_tune.get_embeddings(dataset)
    print(embeddings[:10])
    ```

    Parameters
    ----------
    geneformer_config : GeneformerConfig
        The Geneformer configs to fine-tune, the same as instantiating the standard Geneformer model.
    fine_tuning_head : Literal["classification", "regression"] | HelicalBaseFineTuningHead
        The fine-tuning head that is appended to the model. This can either be a string (options available: "classification", "regression") specifying the task or a custom fine-tuning head inheriting from HelicalBaseFineTuningHead.
    output_size : Optional[int]
        The output size of the fine-tuning model. This is required if the `fine_tuning_head` is a string specified task. For a classification task this is number of unique classes.

    Methods
    -------
    train(train_dataset: Dataset, optimizer: optim, optimizer_params: dict, loss_function: loss, label: str, epochs: int, freeze_layers: int, validation_dataset: Optional[Dataset], lr_scheduler_params: Optional[dict], silent = False)
        Fine-tunes the Geneformer model.
    get_outputs(dataset: Dataset, silent = False)
        Get outputs from the fine-tuned model on the given processed dataset.
    """

    def __init__(
        self,
        geneformer_config: GeneformerConfig,
        fine_tuning_head: (
            Literal["classification", "regression"] | HelicalBaseFineTuningHead
        ),
        output_size: Optional[int] = None,
    ):

        HelicalBaseFineTuningModel.__init__(self, fine_tuning_head, output_size)
        Geneformer.__init__(self, geneformer_config)

        self.fine_tuning_head.set_dim_size(self.config["embsize"])

    def _forward(
        self,
        input_ids: torch.tensor,
        attention_mask_minibatch: torch.tensor,
        original_lengths: torch.tensor,
    ) -> torch.tensor:
        """
        Forward method of the fine-tuning model.

        Parameters
        ----------
        input_ids : torch.tensor
            The input ids to the fine-tuning model.
        attention_mask_minibatch : torch.tensor
            The attention mask for the input tensor.
        original_lengths: torch.tensor
            The original lengths of the inputs without padding

        Returns
        -------
        torch.tensor
            The output tensor of the fine-tuning model.
        """
        outputs = self.model(
            input_ids=input_ids, attention_mask=attention_mask_minibatch
        )
        batch_embeddings = outputs.hidden_states[-1]

        if self.emb_mode == "cls" and self.cls_present:
            batch_embeddings = batch_embeddings[:, 0, :]
        else:
            length = original_lengths
            if self.cls_present:
                batch_embeddings = batch_embeddings[
                    :, 1:, :
                ]  # Get all layers except the cls embs
                if self.eos_present:
                    length -= 2  # length is used for the mean calculation, 2 is subtracted because we have taken both the cls and eos embeddings out
                else:
                    length -= 1  # length is subtracted because just the cls is removed

            batch_embeddings = mean_nonpadding_embs(batch_embeddings, length)

        final = self.fine_tuning_head(batch_embeddings)
        return final

    def train(
        self,
        train_dataset: Dataset,
        optimizer: optim = optim.AdamW,
        optimizer_params: dict = {"lr": 0.0001},
        loss_function: loss = loss.CrossEntropyLoss(),
        label: str = "cell_types",
        epochs: int = 1,
        freeze_layers: int = 2,
        validation_dataset: Optional[Dataset] = None,
        lr_scheduler_params: Optional[dict] = None,
        silent=False,
    ):
        """
        Fine-tunes the Geneformer model for classification tasks.

        Parameters
        ----------
        train_dataset : Dataset
            A helical-processed dataset for fine-tuning.

        optimizer : torch.optim.Optimizer, optional, default=torch.optim.AdamW
            The optimizer to be used for training.

        optimizer_params : dict
            Parameters to be passed to the specified optimizer. This dictionary should *NOT* include model parameters.
            Example:
                optimizer_params = {'lr': 0.0001}

        loss_function : torch.nn.loss, optional, default=torch.nn.CrossEntropyLoss()
            The loss function to be used for training.

        label : str, optional, default="cell_types"
            The column in the dataset containing the training labels. Labels should be stored as unique integers per class.

        epochs : int, optional, default=10
            The number of epochs to train the model.

        freeze_layers : int, optional, default=2
            The number of layers to freeze during training.

        validation_dataset : Dataset, optional, default=None
            A helical-processed dataset used for per-epoch validation. If not specified, no validation will be performed.

        lr_scheduler_params : dict, optional, default=None
            Parameters for the learning rate scheduler from the Transformers [`get_scheduler`](https://huggingface.co/docs/transformers/en/main_classes/optimizer_schedules#transformers.get_scheduler) method. The optimizer should not be included
            in this dictionary, as it will be inferred from the [`optimizer`](https://pytorch.org/docs/main/optim.html) parameter. If not specified, no learning rate scheduler
            will be used.
            Example:
                lr_scheduler_params = {'name': 'linear', 'num_warmup_steps': 0, 'num_training_steps': 5}
        """

        model_input_size = get_model_input_size(self.model)

        _check_for_expected_special_tokens(
            train_dataset,
            self.emb_mode,
            self.cls_present,
            self.eos_present,
            self.tk.gene_token_dict,
        )

        total_batch_length = len(train_dataset)
        # initialise optimizer
        optimizer = optimizer(self.parameters(), **optimizer_params)

        # initialise lr_scheduler
        lr_scheduler = None
        if lr_scheduler_params is not None:
            lr_scheduler = get_scheduler(optimizer=optimizer, **lr_scheduler_params)

        if freeze_layers > 0:
            logger.info(
                f"Freezing the first {freeze_layers} encoder layers of the Geneformer model during fine-tuning."
            )

            frozen_layers = self.model.bert.encoder.layer[:freeze_layers]

            for module in frozen_layers:
                for param in module.parameters():
                    param.requires_grad = False

        self.to(self.device)

        validation_batch_length = 0
        if validation_dataset is not None:
            validation_batch_length = len(validation_dataset)

        logger.info("Starting Fine-Tuning")
        for j in range(epochs):
            training_loop = trange(
                0,
                total_batch_length,
                self.config["batch_size"],
                desc="Fine-Tuning",
                leave=(not silent),
            )
            batch_loss = 0.0
            batches_processed = 0
            for i in training_loop:
                max_range = min(i + self.config["batch_size"], total_batch_length)

                minibatch = train_dataset.select([i for i in range(i, max_range)])
                max_len = int(max(minibatch["length"]))
                minibatch.set_format(type="torch", device=self.device)

                input_data_minibatch = minibatch["input_ids"]
                input_data_minibatch = pad_tensor_list(
                    input_data_minibatch, max_len, self.pad_token_id, model_input_size
                ).to(self.device)

                outputs = self._forward(
                    input_ids=input_data_minibatch,
                    attention_mask_minibatch=gen_attention_mask(minibatch),
                    original_lengths=minibatch["length"],
                )
                loss = loss_function(outputs, minibatch[label])
                loss.backward()
                batch_loss += loss.item()
                batches_processed += 1
                training_loop.set_postfix({"loss": batch_loss / batches_processed})
                training_loop.set_description(f"Fine-Tuning: epoch {j+1}/{epochs}")
                optimizer.step()
                optimizer.zero_grad()

                del outputs
                del minibatch
                del input_data_minibatch
            if lr_scheduler is not None:
                lr_scheduler.step()

            if validation_dataset is not None:
                testing_loop = trange(
                    0,
                    validation_batch_length,
                    self.config["batch_size"],
                    desc="Fine-Tuning Validation",
                    leave=(not silent),
                )
                val_loss = 0.0
                count = 0.0
                for i in testing_loop:
                    max_range = min(
                        i + self.config["batch_size"], validation_batch_length
                    )

                    minibatch = validation_dataset.select(
                        [i for i in range(i, max_range)]
                    )
                    max_len = int(max(minibatch["length"]))
                    minibatch.set_format(type="torch", device=self.device)

                    input_data_minibatch = minibatch["input_ids"]
                    input_data_minibatch = pad_tensor_list(
                        input_data_minibatch,
                        max_len,
                        self.pad_token_id,
                        model_input_size,
                    ).to(self.device)

                    with torch.no_grad():
                        outputs = self._forward(
                            input_ids=input_data_minibatch,
                            attention_mask_minibatch=gen_attention_mask(minibatch),
                            original_lengths=minibatch["length"],
                        )
                    val_loss += loss_function(outputs, minibatch[label]).item()
                    count += 1.0
                    testing_loop.set_postfix({"val_loss": val_loss / count})

                    del outputs
                    del minibatch
                    del input_data_minibatch
        logger.info(f"Fine-Tuning Complete. Epochs: {epochs}")

    def get_outputs(self, dataset: Dataset, silent=False):
        """Predicts the labels for a dataset using the fine-tuned model.

        Parameters
        ----------
        dataset : Dataset
            The processed dataset to generate outputs for.

        Returns
        -------
        np.ndarray
            The predicted labels in the form of a numpy array
        """
        self.model.eval()
        self.fine_tuning_head.eval()

        model_input_size = get_model_input_size(self.model)
        self.to(self.device)

        dataset_length = len(dataset)

        _check_for_expected_special_tokens(
            dataset,
            self.emb_mode,
            self.cls_present,
            self.eos_present,
            self.tk.gene_token_dict,
        )

        output = []
        testing_loop = trange(
            0,
            dataset_length,
            self.config["batch_size"],
            desc="Generating Outputs",
            leave=(not silent),
        )
        for i in testing_loop:
            max_range = min(i + self.config["batch_size"], dataset_length)

            minibatch = dataset.select([i for i in range(i, max_range)])
            max_len = int(max(minibatch["length"]))
            minibatch.set_format(type="torch", device=self.device)

            input_data_minibatch = minibatch["input_ids"]
            input_data_minibatch = pad_tensor_list(
                input_data_minibatch, max_len, self.pad_token_id, model_input_size
            ).to(self.device)

            with torch.no_grad():
                outputs = self._forward(
                    input_ids=input_data_minibatch,
                    attention_mask_minibatch=gen_attention_mask(minibatch),
                    original_lengths=minibatch["length"],
                )
                output.append(outputs.clone().detach())
            del outputs
            del minibatch
            del input_data_minibatch

        return torch.cat(output, dim=0).cpu().numpy()



"""Transcriptformer model implementation."""

import logging
from typing import Literal
import torch
from torch import Tensor, nn
from torch.nn.attention.flex_attention import and_masks, create_block_mask
import pandas as pd

from helical.models.transcriptformer.data.dataclasses import (
    AuxVocab,
    BatchData,
    DataConfig,
    GeneVocab,
    InferenceConfig,
    LossConfig,
    ModelConfig,
)
from helical.models.transcriptformer.model_dir.layers import (
    MLP,
    CountDecoderHead,
    PretrainedEmbeddings,
    TranscriptEncoder,
    mean_embeddings,
)
from helical.models.transcriptformer.model_dir.losses import (
    ZTP_NLL,
    CrossEntropyLoss,
    logit_softcap,
)
from helical.models.transcriptformer.model_dir.masks import (
    causal_mask_factory,
    pad_mask_factory,
)

logger = logging.getLogger(__name__)

NON_GENE_TOKENS = ["unknown", "[PAD]", "[START]", "[END]", "[RD]", "[CELL]", "[MASK]"]


class Transcriptformer(nn.Module):
    """Autoregressive model that predicts the gene tokens and counts of a cell."""

    def __init__(
        self,
        data_config: DataConfig,
        model_config: ModelConfig,
        loss_config: LossConfig,
        inference_config: InferenceConfig = None,
        gene_vocab_dict: dict = None,
        aux_vocab_dict: dict = None,
        emb_matrix: Tensor = None,
        emb_mode: Literal["gene", "cell"] = "cell",
        **kwargs,
    ):
        super().__init__()

        # Store configurations as dataclass objects
        self.data_config = data_config
        self.model_config = model_config
        self.loss_config = loss_config
        self.inference_config = inference_config
        self.emb_mode = emb_mode
        # initialize vocab dicts
        self.gene_vocab_dict = gene_vocab_dict
        self.aux_vocab_dict = aux_vocab_dict

        # Load vocabularies and initialize model components
        self._build_vocabs()
        self._init_gene_embeddings(emb_matrix)
        self._init_model_components()

    def _build_vocabs(self):
        # Load the gene vocabulary
        self.gene_vocab = GeneVocab(
            vocab_dict=self.gene_vocab_dict,
            vocab_size=len(self.gene_vocab_dict),
            pad_idx=self.gene_vocab_dict["[PAD]"],
            start_idx=self.gene_vocab_dict.get("[START]"),
            end_idx=self.gene_vocab_dict.get("[END]"),
            cell_idx=self.gene_vocab_dict.get("[CELL]"),
            embedding_matrix=None,
        )
        # Create reverse dictionary for efficient token to gene name lookup
        self.token_to_gene_dict = {v: k for k, v in self.gene_vocab_dict.items()}

        # Load the auxiliary vocab if provided
        if self.aux_vocab_dict is not None:
            self.model_config.use_aux = True
            self.aux_vocab = AuxVocab(
                vocab_dict=self.aux_vocab_dict,
                vocab_size=sum([len(v) + 1 for v in self.aux_vocab_dict.values()]),
                pad_ids=[
                    self.aux_vocab_dict[k]["unknown"] for k in self.aux_vocab_dict
                ],
                aux_seq_len=len(self.aux_vocab_dict),
            )

    def _init_model_components(self):
        if hasattr(self.model_config, "use_aux"):
            self._init_aux_embeddings()

        # Encoder and heads
        self._init_encoder()
        self._init_count_heads()
        if self.loss_config.gene_id_loss_weight > 0:
            self._init_gene_id_head()

        # Loss initialization
        self._init_nll_loss()
        if self.loss_config.gene_id_loss_weight > 0:
            self._init_gene_id_loss()

        self._init_masks()

    def _init_masks(self):
        self.causal_mask = causal_mask_factory()

    def _init_gene_embeddings(self, emb_matrix: Tensor = None):
        self.gene_embeddings = PretrainedEmbeddings(
            embedding_matrix=emb_matrix,
            output_dim=self.model_config.embed_dim,
            freeze=True,
            normalize=True,
            dropout=self.model_config.dropout,
        )

    def _init_aux_embeddings(self):
        # Initialize auxiliary embeddings based on model_config and vocab
        aux_vocab_dict = self.aux_vocab.vocab_dict
        self.aux_embeddings = nn.ModuleDict(
            {
                key: nn.Embedding(len(vocab), self.model_config.embed_dim)
                for key, vocab in aux_vocab_dict.items()
            }
        )

    def _init_encoder(self):
        # Initialize the transformer encoder using model_config attributes
        self.transformer_encoder = TranscriptEncoder(
            embed_dim=self.model_config.embed_dim,
            num_head=self.model_config.num_heads,
            model_dim=self.model_config.model_dim,
            dropout=self.model_config.dropout,
            nlayers=self.model_config.num_layers,
            activation=self.model_config.activation,
            attn_bias=self.model_config.attn_bias,
            fw_bias=self.model_config.fw_bias,
        )

    def _init_count_heads(self):
        # Head layers
        self.mu = CountDecoderHead(
            model_dim=2
            * self.model_config.embed_dim,  # Double the embedding dim for the skip connection
            link_func=self.model_config.mu_link_fn,
            eps=self.model_config.log_counts_eps,
            dropout=self.model_config.dropout,
            gene_bias_size=self.gene_vocab.vocab_size,  # Gene specific bias
        )

    def _init_gene_id_head(self):
        # initialize gene id head
        self.gene_id_head = MLP(
            self.model_config.embed_dim,
            self.model_config.gene_head_hidden_dim,
            self.gene_vocab.vocab_size,
            dropout=self.model_config.dropout,
        )

    def _init_nll_loss(self):
        self.criterion = ZTP_NLL(
            eps=self.model_config.log_counts_eps,
            softplus_approx=self.loss_config.softplus_approx,
            max_counts=self.data_config.clip_counts,
        )

    def _init_gene_id_loss(self):
        self.gene_id_criterion = CrossEntropyLoss(
            shift_right=False,
            softcap=self.model_config.softcap,
        )

    def _pad_mask(self, gene_tokens, aux_tokens=None, dtype="float"):
        # Create the pad mask
        # False for masked positions
        pad_mask = gene_tokens == self.gene_vocab.pad_idx
        if aux_tokens is not None:
            aux_pad_mask = torch.stack(
                [aux_tokens == pad_id for pad_id in self.aux_vocab.pad_ids], dim=-1
            ).any(dim=-1)
            pad_mask = torch.cat([aux_pad_mask, pad_mask], dim=1)
        # convert mask to float
        if dtype == "float":
            pad_mask = pad_mask.float().masked_fill(pad_mask, float("-inf"))
        elif dtype == "bool":
            pad_mask = ~pad_mask
        pad_mask.requires_grad = False
        return pad_mask

    def get_gene_embeddings(self, batch: BatchData):
        right_shifted_gene_tokens = torch.cat(
            [
                self.gene_vocab.start_idx
                * torch.ones_like(batch.gene_token_indices[:, :1]),
                batch.gene_token_indices[:, :-1],
            ],
            dim=1,
        )

        return self.gene_embeddings(right_shifted_gene_tokens), self.gene_embeddings(
            batch.gene_token_indices
        )

    def get_aux_embeddings(self, aux_token_indices):
        aux_embeddings = []
        for i, aux_embedding in enumerate(self.aux_embeddings.values()):
            aux_embeddings.append(aux_embedding(aux_token_indices[:, i]))
        aux_embeddings = torch.stack(aux_embeddings, dim=1)
        return aux_embeddings

    def _score_mod_factory(self, log_counts, emb_mode=False):
        softcap = self.model_config.softcap

        def score_mod(score, b, h, q_idx, kv_idx):
            bias = log_counts[b, kv_idx] * ((q_idx > kv_idx) | emb_mode)
            score = score + bias
            if softcap > 0:
                score = logit_softcap(score, softcap)
            return score

        return score_mod

    def _get_gene_embeddings(self, transformer_output) -> list[pd.Series]:
        """
        Get the gene embeddings from the transformer output.

        Returns
        -------
            list[pd.Series]: A list of pandas Series, one for each cell, with the gene names as index and the embeddings as values.
        """
        # Create a list where each entry is a cell's gene embeddings Series
        gene_embeddings = []
        gene_output_cpu = transformer_output["gene_embeddings"].detach().cpu().numpy()
        gene_tokens = transformer_output["input_gene_token_indices"]

        for cell_idx in range(gene_output_cpu.shape[0]):
            # Get the actual gene tokens for this cell
            cell_gene_tokens = gene_tokens[cell_idx]
            gene_names = []
            embeddings = []
            for pos_idx in range(gene_output_cpu.shape[1]):
                if not transformer_output["mask"][
                    cell_idx, pos_idx
                ]:  # Only include non-padded positions
                    gene_token = cell_gene_tokens[pos_idx].item()
                    gene_name = self.token_to_gene_dict[gene_token]
                    if gene_name in NON_GENE_TOKENS:
                        continue
                    gene_names.append(gene_name)
                    embeddings.append(gene_output_cpu[cell_idx, pos_idx])
            # Create a pandas Series with gene names as index
            cell_gene_series = pd.Series(embeddings, index=gene_names)
            gene_embeddings.append(cell_gene_series)
        return gene_embeddings

    def forward(
        self,
        batch: BatchData,
        embed: bool = False,
        **kwargs,
    ) -> dict:
        """Forward pass of the model.

        Args:
            batch: BatchData containing gene counts, token indices, and optional auxiliary data
            embed: Whether to compute embeddings
            **kwargs: Additional arguments

        Returns
        -------
            dict: Model output containing:
                - mu: Predicted rate of count distribution
                - mask: Output mask
                - input_counts: Input count data
                - embeddings: Optional cell embeddings
        """
        device = next(self.parameters()).device
        aux_token_indices = batch.aux_token_indices.to(device)
        gene_token_indices = batch.gene_token_indices.to(device)
        gene_counts = batch.gene_counts.to(device)

        # append the start token to the gene_tokens
        # drop the last token from the gene_tokens to keep the same length as the counts
        right_shifted_gene_tokens = torch.cat(
            [
                self.gene_vocab.start_idx * torch.ones_like(gene_token_indices[:, :1]),
                gene_token_indices[:, :-1],
            ],
            dim=1,
        )
        right_shifted_counts = torch.cat(
            [torch.ones_like(gene_counts[:, :1]), gene_counts[:, :-1]], dim=1
        )

        # Embed the gene_tokens
        right_shifted_gene_embeddings, gene_embeddings = self.get_gene_embeddings(batch)

        # Add auxiliary tokens to the embeddings
        if aux_token_indices is not None:
            aux_embeddings = self.get_aux_embeddings(aux_token_indices)
            right_shifted_gene_embeddings = torch.cat(
                [aux_embeddings, right_shifted_gene_embeddings], dim=1
            )
            right_shifted_counts = torch.cat(
                [torch.ones_like(aux_token_indices), right_shifted_counts], dim=1
            )

        # Compile the block mask
        pad_mask = self._pad_mask(
            right_shifted_gene_tokens, aux_token_indices, dtype="bool"
        )
        mask_mod = and_masks(pad_mask_factory(pad_mask), self.causal_mask)

        block_mask = create_block_mask(
            mask_mod,
            right_shifted_gene_embeddings.shape[0],
            H=None,
            Q_LEN=right_shifted_gene_embeddings.shape[1],
            KV_LEN=right_shifted_gene_embeddings.shape[1],
            BLOCK_SIZE=self.model_config.block_len,
            _compile=True,
        )

        # Score mode
        log_counts = torch.log1p(
            right_shifted_counts + self.model_config.log_counts_eps
        )
        score_mod = self._score_mod_factory(log_counts, emb_mode=embed)

        # Apply the transformer encoder
        transformer_output = self.transformer_encoder(
            x=right_shifted_gene_embeddings,
            score_mod=score_mod,
            block_mask=block_mask,
        )

        # Extract the auxiliary features from the output
        if aux_token_indices is not None:
            # Determine the number of auxiliary tokens
            num_aux_tokens = aux_token_indices.shape[1]

            # Separate the auxiliary output and gene output from the concatenated embeddings
            gene_output = transformer_output[:, num_aux_tokens:, :]

            # Adjust the pad mask to exclude the auxiliary tokens
            pad_mask = pad_mask[:, num_aux_tokens:]
        else:
            gene_output = transformer_output

        result = {}

        if embed:
            result["cell_embeddings"] = (
                mean_embeddings(gene_output, pad_mask).detach().cpu()
            )
            result["gene_embeddings"] = gene_output.detach().cpu()

        # Append the end token to the gene_tokens
        result["input_gene_token_indices"] = torch.cat(
            [
                gene_token_indices[:, :-1],
                self.gene_vocab.end_idx * torch.ones_like(gene_token_indices[:, :1]),
            ],
            dim=1,
        )

        # set last count to zero
        result["input_counts"] = torch.cat(
            [gene_counts[:, :-1], torch.ones_like(gene_counts[:, :1])], dim=1
        )

        # concatenate gene_output with input embeddings
        # this allows the model to condition on the input gene tokens
        conditioned_output = torch.cat([gene_output, gene_embeddings], dim=-1)

        # Calculate the mu parameter
        result["mu"] = self.mu(
            gene_output=conditioned_output,
            gene_tokens=gene_token_indices,
            mask=self._pad_mask(gene_token_indices, dtype="float"),
        )
        if self.model_config.mu_link_fn == "softmax":
            # avoid in-place operations with tensor involved in the backward pass
            result["mu"] = result["mu"] * gene_counts.sum(dim=1, keepdim=True)

        # Mask the output
        # convention for loss functions is true = masked
        result["mask"] = ~self._pad_mask(gene_token_indices, dtype="bool")

        # Calculate the gene id loss
        if self.loss_config.gene_id_loss_weight > 0:
            gene_logit = self.gene_id_head(gene_output)
            result["gene_logit"] = gene_logit

        return result

    @torch.no_grad()
    def inference(self, batch: BatchData):
        """Run inference on a batch of data.

        Args:
            batch: BatchData containing input data

        Returns
        -------
            Dictionary containing inference results
        """
        # Flex Attn requires batch size to be a multiple of BLOCK_SIZE
        batch_size = batch.gene_counts.shape[0]

        if batch_size > self.inference_config.batch_size:
            raise ValueError(
                f"Batch size {batch_size} exceeds model batch size {self.inference_config.batch_size}"
            )

        pad_rows = self.inference_config.batch_size - batch_size

        resized_batch = BatchData(
            **{
                field_name: (
                    self._resize_data(field_value, self.inference_config.batch_size)
                )
                for field_name, field_value in batch.__dict__.items()
            }
        )

        transformer_output = self.forward(batch=resized_batch, embed=True)

        if pad_rows > 0:
            # Remove the last pad_rows rows from the batch
            for key, value in transformer_output.items():
                if isinstance(value, torch.Tensor):
                    transformer_output[key] = value[:-pad_rows]
                    assert len(transformer_output[key]) == batch_size

        if "llh" in self.inference_config.output_keys:
            llh = (
                self.criterion(**transformer_output, eval_mode=not self.training)
                .detach()
                .cpu()
            )
            transformer_output["llh"] = llh

        if "gene_llh" in self.inference_config.output_keys:
            self.gene_id_criterion.reduction = None
            gene_llh = (
                self.gene_id_criterion(
                    logits=transformer_output["gene_logit"],
                    input_ids=transformer_output["input_gene_token_indices"],
                    mask=transformer_output["mask"],
                )
                .detach()
                .cpu()
            )
            transformer_output["gene_llh"] = gene_llh

        results = {}
        results["obs"] = batch.obs
        results.update(
            {key: transformer_output[key] for key in self.inference_config.output_keys}
        )

        if self.emb_mode == "gene":
            results["gene_embeddings"] = self._get_gene_embeddings(transformer_output)
        elif self.emb_mode == "cell":
            results["cell_embeddings"] = (
                transformer_output["cell_embeddings"].detach().cpu()
            )
        else:
            raise ValueError(f"Invalid embedding mode: {self.emb_mode}")

        return results

    def _resize_data(self, data, config_batch_size):
        """
        Resize batch to match model batch_size.

        Necessary for inference when batch_size is not a multiple of model batch_size.
        """
        if not isinstance(data, torch.Tensor):
            return data

        batch_size = data.shape[0]
        if batch_size != config_batch_size:
            data = self._pad_rows(
                config_batch_size - batch_size,
                data,
                "int" if data.dtype == torch.int64 else "float",
            )
        return data

    def _pad_rows(self, pad_len, x, dtype="int"):
        # Create padding with same shape as x except for first dimension
        pad_shape = (pad_len,) + x.shape[1:]
        pad = torch.zeros(pad_shape, device=x.device)
        if dtype == "int":
            pad = pad.int()
        elif dtype == "float":
            pad = pad.float()
        else:
            raise ValueError(f"Invalid padding type: {dtype}")
        return torch.cat([x, pad], dim=0)

    def predict_step(self, batch, batch_idx):
        assert (
            self.inference_config.output_keys
        ), "output_keys must be set in inference_config"
        return self.inference(batch)



import pytest
import torch
from helical.models.hyena_dna import HyenaDNAConfig, HyenaDNAFineTuningModel
from helical.models.fine_tune.fine_tuning_heads import ClassificationHead
import os

class TestHyenaDNAFineTuning:
    @pytest.fixture(params=["hyenadna-tiny-1k-seqlen", "hyenadna-tiny-1k-seqlen-d256"])
    def hyenaDNAFineTune(self, request):
        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        config = HyenaDNAConfig(
            model_name=request.param, batch_size=1, device=self.device
        )
        return HyenaDNAFineTuningModel(
            hyena_config=config, fine_tuning_head="classification", output_size=1
        )

    @pytest.fixture
    def mock_data(self, hyenaDNAFineTune):
        input_sequences = ["AAAA", "CCCC", "TTTT", "ACGT", "ACGN", "ANNT"]
        labels = [0, 0, 0, 0, 0, 0]
        tokenized_sequences = hyenaDNAFineTune.process_data(input_sequences)
        return tokenized_sequences, labels

    def test_output_dimensionality_of_fine_tuned_model(
        self, hyenaDNAFineTune, mock_data
    ):
        input_sequences, labels = mock_data
        hyenaDNAFineTune.train(
            train_dataset=input_sequences,
            train_labels=labels,
            validation_dataset=input_sequences,
            validation_labels=labels,
        )
        outputs = hyenaDNAFineTune.get_outputs(input_sequences)
        assert outputs.shape == (len(input_sequences), 1)

    def test_save_and_load_model(self, hyenaDNAFineTune):

        try:
            hyenaDNAFineTune.save_model("./hyena_dna_fine_tuned_model.pt")
            hyenaDNAFineTune.load_model("./hyena_dna_fine_tuned_model.pt")
            assert not hyenaDNAFineTune.model.training, "Model should be in eval mode"
            assert not hyenaDNAFineTune.fine_tuning_head.training, "Fine-tuning head should be in eval mode"
            assert hyenaDNAFineTune.model is not None
            assert hyenaDNAFineTune.fine_tuning_head.output_size == 1, "Output size should be 1"
            assert type(hyenaDNAFineTune.fine_tuning_head) == ClassificationHead, "Fine-tuning head should be a ClassificationHead"
            assert hyenaDNAFineTune.model.state_dict() is not None
        finally:
            if os.path.exists("./hyena_dna_fine_tuned_model.pt"):
                os.remove("./hyena_dna_fine_tuned_model.pt")


import os
from typing import Union
import numpy as np
import torch

PathLike = Union[str, os.PathLike]


class Dataset(torch.utils.data.Dataset):
    def __init__(self, count_matrix, gene_ids, vocab, model_configs, batch_ids=None):
        self.count_matrix = count_matrix
        self.gene_ids = gene_ids
        self.batch_ids = batch_ids
        self.vocab = vocab
        self.model_configs = model_configs

    def __len__(self):
        return len(self.count_matrix)

    def __getitem__(self, idx):
        row = self.count_matrix[idx]
        nonzero_idx = np.nonzero(row)[0]
        values = row[nonzero_idx]
        genes = self.gene_ids[nonzero_idx]
        # append <cls> token at the beginning
        genes = np.insert(genes, 0, self.vocab["<cls>"])
        values = np.insert(values, 0, self.model_configs["pad_value"])
        genes = torch.from_numpy(genes).long()
        values = torch.from_numpy(values).float()
        output = {
            "id": idx,
            "genes": genes,
            "expressions": values,
        }
        if self.batch_ids is not None:
            output["batch_labels"] = self.batch_ids[idx]
        return output



import logging
import numpy as np
from anndata import AnnData
from torch.utils.data import DataLoader
from pathlib import Path
import scanpy as sc
from tqdm import tqdm
import torch
from accelerate import Accelerator
from helical.models.uce.uce_config import UCEConfig
from helical.models.base_models import HelicalRNAModel
from helical.models.uce.uce_utils import (
    get_ESM2_embeddings,
    get_positions,
    get_protein_embeddings_idxs,
    load_model,
    prepare_expression_counts_file,
)
from helical.utils.downloader import Downloader
from helical.models.uce.uce_dataset import UCEDataset
from helical.models.uce.gene_embeddings import load_gene_embeddings_adata

LOGGER = logging.getLogger(__name__)


class UCE(HelicalRNAModel):
    """Universal Cell Embedding Model. This model reads in single-cell RNA-seq data and outputs gene embeddings.
    This model particularly uses protein-embeddings generated by ESM2.
    Currently we support human and macaque species but you can add your own species by providing the protein embeddings.

    Example
    -------
    ```python
    from helical.models.uce import UCE, UCEConfig
    from datasets import load_dataset
    from helical.utils import get_anndata_from_hf_dataset
    import anndata as ad

    configurer=UCEConfig(batch_size=10)
    uce = UCE(configurer=configurer)

    hf_dataset = load_dataset("helical-ai/yolksac_human",split="train[:25%]", trust_remote_code=True, download_mode="reuse_cache_if_exists")
    ann_data = get_anndata_from_hf_dataset(hf_dataset)

    dataset = uce.process_data(ann_data[:100])

    embeddings = uce.get_embeddings(dataset)

    print("UCE embeddings ", embeddings[:10])
    ```

    Parameters
    ----------
    configurer : UCEConfig, optional, default=default_configurer
        The model configuration.

    Notes
    -----
    The Universal Cell Embedding Paper has been published on [bioRxiv](https://www.biorxiv.org/content/10.1101/2023.11.28.568918v1) and it is built on top of SATURN published in [Nature](https://www.nature.com/articles/s41592-024-02191-z).
    """

    default_configurer = UCEConfig()

    def __init__(self, configurer: UCEConfig = default_configurer) -> None:
        super().__init__()
        self.config = configurer.config

        downloader = Downloader()
        for file in self.config["list_of_files_to_download"]:
            downloader.download_via_name(file)

        self.model_dir = self.config["model_path"].parent
        self.device = self.config["device"]
        self.embeddings = get_ESM2_embeddings(
            self.config["token_file_path"], self.config["token_dim"]
        )
        self.model = load_model(self.config["model_path"], self.config, self.embeddings)
        self.model = self.model.eval().to(self.device)

        if self.config["accelerator"] or self.device == "cuda":
            self.accelerator = Accelerator(
                project_dir=self.model_dir
            )  # , cpu=self.config["accelerator"]["cpu"])
            self.model = self.accelerator.prepare(self.model)
        else:
            self.accelerator = None

        LOGGER.info(f"Model finished initializing.")
        mode = "training" if self.model.training else "eval"
        LOGGER.info(
            f"'{self.config['model_name']}' model is in '{mode}' mode, on device '{next(self.model.parameters()).device.type}'."
        )

    def process_data(
        self,
        adata: AnnData,
        gene_names: str = "index",
        name="test",
        filter_genes_min_cell: int = None,
        use_raw_counts: bool = True,
    ) -> UCEDataset:
        """
        Processes the data for the Universal Cell Embedding model.

        Parameters
        ----------
        adata : AnnData
            The AnnData object containing the data to be processed.
            The UCE model requires the gene expression data as input and the gene symbols
            as variable names (i.e., `adata.var_names`).
        gene_names : str, optional, default="index"
            The name of the column in the AnnData object that contains the gene symbols.
            By default, the index of the AnnData object is used. If another column is specified,
            that column will be set as the index of the AnnData object.
        name : str, optional, default="test"
            The name of the dataset. Needed for when slicing AnnData objects for train and validation datasets.
        filter_genes_min_cell : int, optional, default=None
            Filter threshold that defines how many times a gene should occur in all the cells.
        use_raw_counts : bool, optional, default=True
            Whether to use raw counts or not.

        Returns
        -------
        UCEDataset
            An object that inherits from the `Dataset` class.
        """

        LOGGER.info(f"Processing data for UCE.")
        self.ensure_rna_data_validity(adata, gene_names, use_raw_counts)

        if gene_names != "index":
            adata.var.index = adata.var[gene_names]

        files_config = {
            "spec_chrom_csv_path": self.model_dir / "species_chrom.csv",
            "protein_embeddings_dir": self.model_dir / "protein_embeddings/",
            "offset_pkl_path": self.model_dir / "species_offsets.pkl",
        }

        ## TODO : Remove double downloads. This is required since metaflow might not have stored the files in the right location and the files might have dissapeared. The downloader should check if the file already exists.
        downloader = Downloader()
        for file in self.config["list_of_files_to_download"]:
            if "torch" in file:
                continue
            downloader.download_via_name(file)

        if filter_genes_min_cell is not None:
            sc.pp.filter_genes(adata, min_cells=filter_genes_min_cell)
            # sc.pp.filter_cells(ad, min_genes=25)
        ##Filtering out the Expression Data That we do not have in the protein embeddings
        filtered_adata, species_to_all_gene_symbols = load_gene_embeddings_adata(
            adata=adata,
            species=[self.config["species"]],
            embedding_model=self.config["gene_embedding_model"],
            embeddings_path=Path(files_config["protein_embeddings_dir"]),
        )
        # TODO: What about hv_genes? See orig.
        gene_expression = adata.X.toarray()

        name = name
        gene_expression_folder_path = "./"
        prepare_expression_counts_file(
            gene_expression, name, gene_expression_folder_path
        )

        # shapes dictionary
        num_cells = filtered_adata.X.shape[0]
        num_genes = filtered_adata.X.shape[1]
        shapes_dict = {name: (num_cells, num_genes)}

        pe_row_idxs = get_protein_embeddings_idxs(
            files_config["offset_pkl_path"],
            self.config["species"],
            species_to_all_gene_symbols,
            filtered_adata,
        )
        dataset_chroms, dataset_start = get_positions(
            Path(files_config["spec_chrom_csv_path"]),
            self.config["species"],
            filtered_adata,
        )

        if not (
            len(dataset_chroms)
            == len(dataset_start)
            == num_genes
            == pe_row_idxs.shape[0]
        ):
            LOGGER.error(
                f"Invalid input dimensions for the UCEDataset! "
                f"dataset_chroms: {len(dataset_chroms)}, "
                f"dataset_start: {len(dataset_start)}, "
                f"num_genes: {num_genes}, "
                f"pe_row_idxs.shape[0]: {pe_row_idxs.shape[0]}"
            )
            raise AssertionError

        dataset = UCEDataset(
            sorted_dataset_names=[name],
            shapes_dict=shapes_dict,
            model_config=self.config,
            expression_counts_path=gene_expression_folder_path,
            dataset_to_protein_embeddings=pe_row_idxs,
            datasets_to_chroms=dataset_chroms,
            datasets_to_starts=dataset_start,
        )
        LOGGER.info(f"Successfully processed the data for UCE.")
        return dataset

    def get_embeddings(self, dataset: UCEDataset) -> np.array:
        """Gets the gene embeddings from the UCE model

        Parameters
        ----------
        dataset : UCEDataSet
            The Dataset object containing the processed data

        Returns
        -------
        np.ndarray
            The gene embeddings in the form of a numpy array
        """
        LOGGER.info(f"Started getting embeddings:")

        batch_size = self.config["batch_size"]
        dataloader = DataLoader(
            dataset,
            batch_size=batch_size,
            shuffle=False,
            collate_fn=dataset.collator_fn,
            num_workers=0,
        )

        if self.accelerator is not None:
            dataloader = self.accelerator.prepare(dataloader)

        # disable progress bar if not the main process
        if self.accelerator is not None:
            pbar = tqdm(dataloader, disable=not self.accelerator.is_local_main_process)
        else:
            pbar = tqdm(dataloader)

        dataset_embeds = []

        # disabling gradient calculation for inference
        with torch.no_grad():
            for batch in pbar:
                batch_sentences, mask, idxs = batch[0], batch[1], batch[2]
                batch_sentences = batch_sentences.permute(1, 0)
                if self.config["multi_gpu"]:
                    batch_sentences = self.model.module.pe_embedding(
                        batch_sentences.long()
                    )
                else:
                    batch_sentences = self.model.pe_embedding(batch_sentences.long())
                batch_sentences = torch.nn.functional.normalize(
                    batch_sentences, dim=2
                )  # normalize token outputs
                _, embedding = self.model.forward(batch_sentences, mask=mask)

                # Fix for duplicates in last batch
                if self.accelerator is not None:
                    self.accelerator.wait_for_everyone()
                    embeddings = self.accelerator.gather_for_metrics((embedding))
                    if self.accelerator.is_main_process:
                        dataset_embeds.append(embeddings.detach().cpu().numpy())
                else:
                    dataset_embeds.append(embedding.detach().cpu().numpy())
        embeddings = np.vstack(dataset_embeds)

        LOGGER.info(f"Finished getting embeddings.")
        return embeddings
